#include "burp.h"
#include "msg.h"
#include "prog.h"
#include "asyncio.h"

int send_msg_fp(FILE *fp, char cmd, const char *buf, size_t s)
{
	fprintf(fp, "%c%04X%s\n", cmd, (unsigned int)s, buf);
	return 0;
}

int send_msg_zp(gzFile zp, char cmd, const char *buf, size_t s)
{
	gzprintf(zp, "%c%04X%s\n", cmd, s, buf);
	return 0;
}

int transfer_file_in(BFILE *bfd, FILE *fp, gzFile zp)
{
	char cmd;
	char *buf=NULL;
	size_t len=0;
	int quit=0;
	int ret=-1;
//logp("in transfer file in\n");

	while(!quit)
	{
		if(async_read(&cmd, &buf, &len)) return -1;

//		logp("transfer in: %c\n", cmd);
		switch(cmd)
		{
			case 'a': // append
				if(!fp && !zp && !bfd)
				{
					logp("given append, but no file to write to\n");
					async_write_str('e', "append with no file");
					quit++; ret=-1;
				}
				else
				{
//					logp("want to write: %d\n", len);
#ifdef HAVE_WIN32
					if((ret=bwrite(bfd, buf, len))<=0)
					{
						logp("error when appending: %d\n", ret);
						quit++; ret=-1;
					}
#else
					if((zp
					    && (ret=gzwrite(zp, buf, len))<=0)
					  || (fp
					    && (ret=fwrite(buf, 1, len, fp))<=0))
					{
						logp("error when appending: %d\n", ret);
						async_write_str('e', "write failed");
						quit++; ret=-1;
					}
#endif
				}
				break;
			case 'x': // finish up
				quit++;
				ret=0;
				break;
			default:
				logp("unknown append cmd: %c\n", cmd);
				quit++;
				ret=-1;
				break;
		}
		if(buf) free(buf);
		buf=NULL;
	}
	if(ret)
		logp("transfer file in returning: %d\n", ret);
	return ret;
}

#define ZCHUNK 16000

int transfer_gzfile_in(BFILE *bfd, FILE *fp, gzFile zp)
{
	char cmd;
	char *buf=NULL;
	size_t len=0;
	int quit=0;
	int ret=-1;
	int zret=Z_OK;
	unsigned char out[ZCHUNK];
	unsigned have=0;
//logp("in transfer gzfile_in\n");

	z_stream zstrm;

	zstrm.zalloc=Z_NULL;
	zstrm.zfree=Z_NULL;
	zstrm.opaque=Z_NULL;
	zstrm.avail_in=0;
	zstrm.next_in=Z_NULL;

	if(inflateInit2(&zstrm, (15+16)))
	{
		logp("unable to init inflate\n");
		return -1;
	}

	while(!quit)
	{
		if(async_read(&cmd, &buf, &len)) return -1;

		//logp("transfer in: %c:%s\n", cmd, buf);
		switch(cmd)
		{
			case 'a': // append
				if(!fp && !zp && !bfd)
				{
					logp("given append, but no file to write to\n");
					async_write_str('e', "append with no file");
					quit++; ret=-1;
				}
				else
				{
					zstrm.avail_in=len;
					zstrm.next_in=(unsigned char *)buf;
					//logp("want to write: %d\n", len);

					do
					{
						zstrm.avail_out=ZCHUNK;
						zstrm.next_out=out;
						zret=inflate(&zstrm, Z_NO_FLUSH);
						switch(zret)
						{
							case Z_NEED_DICT:
							  zret=Z_DATA_ERROR;
							case Z_DATA_ERROR:
							case Z_MEM_ERROR:
							  logp("zstrm inflate error: %d\n", zret);
							  inflateEnd(&zstrm);
							  quit++;
							  ret=-1;
							  continue;
						}
						have=ZCHUNK-zstrm.avail_out;
						if(!have) continue;
#ifdef HAVE_WIN32
						if((ret=bwrite(bfd, out, have))<=0)
						{
							logp("error when appending: %d\n", ret);
							quit++; ret=-1;
						}
#else
						if((zp
						    && (ret=gzwrite(zp,
							out, have))<=0)
						  || (fp
						    && (ret=fwrite(out,
							1, have, fp))<=0))
						{
							logp("error when appending: %d\n", ret);
							async_write_str('e', "write failed");
							quit++; ret=-1;
						}
#endif
					} while(!zstrm.avail_out);
				}
				break;
			case 'x': // finish up
				quit++;
				ret=0;
				break;
			default:
				logp("unknown append cmd: %c\n", cmd);
				quit++;
				ret=-1;
				break;
		}
		if(buf) free(buf);
		buf=NULL;
	}
	if(ret) logp("transfer file returning: %d\n", ret);
	return ret;
}

FILE *open_file(const char *fname, const char *mode)
{
	FILE *fp=NULL;

	if(!(fp=fopen(fname, mode)))
	{
		logp("could not open %s: %s\n", fname, strerror(errno));
		return NULL; 
	}
	return fp;
}

gzFile gzopen_file(const char *fname, const char *mode)
{
	gzFile fp=NULL;

	if(!(fp=gzopen(fname, mode)))
	{
		logp("could not open %s: %s\n", fname, strerror(errno));
		return NULL; 
	}
	return fp;
}
