/*= -*- c-basic-offset: 4; indent-tabs-mode: nil; -*-
 *
 * librsync -- the library for network deltas
 * $Id: buf.c,v 1.22 2003/12/16 00:10:55 abo Exp $
 * 
 * Copyright (C) 2000, 2001 by Martin Pool <mbp@samba.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

                              /*
                               | Pick a window, Jimmy, you're leaving.
                               |   -- Martin Schwenke, regularly
                               */


/*
 * buf.c -- Buffers that map between stdio file streams and librsync
 * streams.  As the stream consumes input and produces output, it is
 * refilled from appropriate input and output FILEs.  A dynamically
 * allocated buffer of configurable size is used as an intermediary.
 *
 * TODO: Perhaps be more efficient by filling the buffer on every call
 * even if not yet completely empty.  Check that it's really our
 * buffer, and shuffle remaining data down to the front.
 *
 * TODO: Perhaps expose a routine for shuffling the buffers.
 */


#include "burp.h"
#include "prog.h"
#include "asyncio.h"
#include "rs_buf.h"
#include "zlibio.h"
#include <assert.h>

/* use fseeko instead of fseek for long file support if we have it */
#ifdef HAVE_FSEEKO
#define fseek fseeko
#endif

/**
 * File IO buffer sizes.
 */
//int rs_inbuflen = 16000, rs_outbuflen = 16000;


/*
 * Allocate and zero-fill an instance of TYPE.
 */
#define rs_alloc_struct(type)                           \
        ((type *) rs_alloc_struct0(sizeof(type), #type))

void *
rs_alloc_struct0(size_t size, char const *name)
{
    void           *p;

    if (!(p = malloc(size))) {
        logp("couldn't allocate instance of %s", name);
    }
    else memset(p, 0, size);
    return p;
}

void *
rs_alloc(size_t size, char const *name)
{
    void           *p;

    if (!(p = malloc(size))) {
        logp("couldn't allocate instance of %s", name);
    }

    return p;
}

#define ZCHUNK	16000

rs_filebuf_t *rs_filebuf_new(BFILE *bfd, FILE *fp, gzFile zp, int fd, enum ztype z,
	size_t buf_len) 
{
    rs_filebuf_t *pf=NULL;
    z_stream *zstrm=NULL;
    unsigned char *zbuf=NULL;
    if(!(pf=rs_alloc_struct(rs_filebuf_t))) return NULL;

    if(z!=ZTYPE_NONE)
    {
	if(!(zstrm=(z_stream *)rs_alloc(sizeof(z_stream), "zstream"))
	  || !(zbuf=(unsigned char *)rs_alloc(ZCHUNK, "zbuf")))
	{
		free(pf);
		return NULL;
	}

	zstrm->zalloc=Z_NULL;
	zstrm->zfree=Z_NULL;
	zstrm->opaque=Z_NULL;
	zstrm->avail_in=0;
	zstrm->next_in=Z_NULL;

	if(z!=ZTYPE_NONE) zstrm->avail_out=buf_len;

	if((z==ZTYPE_DEFLATE
	  && deflateInit2(zstrm, Z_BEST_COMPRESSION, Z_DEFLATED, (15+16),
		8, Z_DEFAULT_STRATEGY)!=Z_OK)
	 || (z==ZTYPE_INFLATE
	  && inflateInit2(zstrm, (15+16))))
	{
		free(pf);
		if(zstrm) free(zstrm);
		return NULL;
	}
    }

    if(!(pf->buf=(char *)rs_alloc(buf_len, "file buffer")))
    {
	free(pf);
	if(zstrm) free(zstrm);
	return NULL;
    }
    pf->buf_len=buf_len;
    pf->fp=fp;
    pf->zp=zp;
    pf->fd=fd;
    pf->bfd=bfd;
    pf->zstrm=zstrm;
    pf->z=z;
    pf->zbuf=zbuf;
    pf->zhave=0;

    return pf;
}

void rs_filebuf_free(rs_filebuf_t *fb) 
{
	if(fb->buf) free(fb->buf);
	if(fb->zbuf) free(fb->zbuf);
	if(fb->z==ZTYPE_DEFLATE && fb->zstrm)
	{
		deflateEnd(fb->zstrm);
		free(fb->zstrm);
	}
	else if(fb->z==ZTYPE_INFLATE && fb->zstrm)
	{
		inflateEnd(fb->zstrm);
		free(fb->zstrm);
	}
	memset(fb, 0, sizeof(*fb));
        free(fb);
	fb=NULL;
}

rs_result zlib_fd_inflate(int fd, rs_filebuf_t *fb, rs_buffers_t *buf);

/*
 * If the stream has no more data available, read some from F into
 * BUF, and let the stream use that.  On return, SEEN_EOF is true if
 * the end of file has passed into the stream.
 */
rs_result rs_infilebuf_fill(rs_job_t *job, rs_buffers_t *buf,
                            void *opaque)
{
    int                     len=0;
    rs_filebuf_t            *fb = (rs_filebuf_t *) opaque;
    gzFile                  zp = fb->zp;
    FILE                    *fp = fb->fp;
    int fd=fb->fd;

//logp("rs_infilebuf_fill\n");

    /* This is only allowed if either the buf has no input buffer
     * yet, or that buffer could possibly be BUF. */
    if (buf->next_in != NULL) {
	//logp("infilebuf avail_in %d buf_len %d\n", buf->avail_in, fb->buf_len);
        assert(buf->avail_in <= fb->buf_len);
        assert(buf->next_in >= fb->buf);
        assert(buf->next_in <= fb->buf + fb->buf_len);
    } else {
        assert(buf->avail_in == 0);
    }

    if (buf->eof_in)
    {
        return RS_DONE;
    }

    if (buf->avail_in)
        /* Still some data remaining.  Perhaps we should read
           anyhow? */
        return RS_DONE;

    if(fd>=0)
    {
	char rcmd;
	size_t rlen;
	char *rbuf=NULL;

	if(fb->z==ZTYPE_INFLATE && fd>0)
	{
		rs_result r=zlib_fd_inflate(fd, fb, buf);
		//logp("buf->avail_in: %d\n", buf->avail_in);
/*
		FILE *rp=NULL;
		char p[64]="";
		snprintf(p, sizeof(p), "/tmp/avail_in.sig");
		rp=fopen(p, "w");
		fwrite(buf->next_in, 1, buf->avail_in, rp);
		fclose(rp);
*/
		return r;
	}

	if(async_read(&rcmd, &rbuf, &rlen)) return RS_IO_ERROR;
	if(rcmd=='a')
	{
		//logp("got 'a' in fd infilebuf: %d\n", rlen);
		memcpy(fb->buf, rbuf, rlen);
		len=rlen;
		free(rbuf);
	}
	else if(rcmd=='x')
	{
		free(rbuf);
		//logp("got 'x' in fd infilebuf\n");
		buf->eof_in=1;
		return RS_DONE;
	}
	else
	{
		logp("unexpected cmd in rs_infilebuf_fill: %c:%s\n", rcmd, rbuf);
		free(rbuf);
		return RS_IO_ERROR;
	}
    }
#ifdef HAVE_WIN32
    else if(fb->bfd)
    {
	len=bread(fb->bfd, fb->buf, fb->buf_len);
	if(len==0)
	{
		//logp("bread: eof\n");
		buf->eof_in=1;
		return RS_DONE;
	}
	else if(len<0)
	{
		logp("rs_infilebuf_fill: error in bread\n");
		return RS_IO_ERROR;
	}
	//logp("bread: ok: %d\n", len);
    }
#endif
    else if(fp)
    {
	    len = fread(fb->buf, 1, fb->buf_len, fp);
//logp("fread: %d\n", len);
	    if (len <= 0) {
		/* This will happen if file size is a multiple of input block len
		 */
		if (feof(fp)) {
		    buf->eof_in=1;
		    return RS_DONE;
		} else {
		    logp("got return %d when trying to read\n", len);
		    return RS_IO_ERROR;
		}
	    }
    }
    else if(zp)
    {
	    len = gzread(zp, fb->buf, fb->buf_len);
//logp("gzread: %d\n", len);
	    if (len <= 0) {
		/* This will happen if file size is a multiple of input block len
		 */
		if (gzeof(zp)) {
		    buf->eof_in=1;
		    return RS_DONE;
		} else {
		    logp("got return %d when trying to read\n", len);
		    return RS_IO_ERROR;
		}
	    }
	    if (len < (int)fb->buf_len && gzeof(fp)) {
		buf->eof_in=1;
		return RS_DONE;
	    }
    }

    buf->avail_in = len;
    buf->next_in = fb->buf;

    return RS_DONE;
}

rs_result zlib_fd_deflate(rs_filebuf_t *fb, size_t *wlen, int flush);

/*
 * The buf is already using BUF for an output buffer, and probably
 * contains some buffered output now.  Write this out to F, and reset
 * the buffer cursor.
 */
rs_result rs_outfilebuf_drain(rs_job_t *job, rs_buffers_t *buf, void *opaque)
{
    rs_filebuf_t *fb = (rs_filebuf_t *) opaque;
    FILE *fp = fb->fp;
    gzFile zp = fb->zp;
    int fd = fb->fd;
    size_t wlen;

//logp("in rs_outfilebuf_drain\n");

    /* This is only allowed if either the buf has no output buffer
     * yet, or that buffer could possibly be BUF. */
    if(!buf->next_out)
    {
        assert(buf->avail_out == 0);
        buf->next_out = fb->buf;
        buf->avail_out = fb->buf_len;
        return RS_DONE;
    }
        
    assert(buf->avail_out <= fb->buf_len);
    assert(buf->next_out >= fb->buf);
    assert(buf->next_out <= fb->buf + fb->buf_len);

    if((wlen=buf->next_out-fb->buf)>0)
    {
	//logp("wlen: %d\n", wlen);
	if(fb->z==ZTYPE_DEFLATE && fd>0)
	{
		rs_result res;
		//logp("about to deflate\n");
		res=zlib_fd_deflate(fb, &wlen, Z_NO_FLUSH);
		//logp("result: %d\n", res);
		if(res==RS_BLOCKED || res==RS_RUNNING)
		{
			// stop the rsync stuff from reading more.
			buf->next_out = fb->buf;
			buf->avail_out = 0;
			//logp("zlib out return BLOCKED\n");
			return res;
		}
		else if(res!=RS_DONE) return res;
	}
	else if(fd>0)
	{
		if(async_append_all_to_write_buffer('a', fb->buf, &wlen))
		{
			// stop the rsync stuff from reading more.
			buf->next_out = fb->buf;
			buf->avail_out = 0;
			logp("out return BLOCKED\n");
			return RS_BLOCKED;
		}
	}
	else
	{
		size_t result=0;
		if(fp) result=fwrite(fb->buf, 1, wlen, fp);
		else if(zp) result=gzwrite(zp, fb->buf, wlen);
		if(wlen!=result)
		{
		    logp("error draining buf to file: %s",
			     strerror(errno));
		    return RS_IO_ERROR;
		}
	}
    }

    buf->next_out = fb->buf;
    buf->avail_out = fb->buf_len;
        
    return RS_DONE;
}

rs_result do_rs_run(rs_job_t *job, BFILE *bfd,
	FILE *in_file, FILE *out_file,
	gzFile in_zfile, gzFile out_zfile, int infd, int outfd,
	enum ztype inz, enum ztype outz)
{
	rs_buffers_t buf;
	rs_result result;
	rs_filebuf_t *in_fb=NULL;
	rs_filebuf_t *out_fb=NULL;

	if(in_file && infd>=0)
	{
		logp("do not specify both input file and input fd in do_rs_run()\n");
		return RS_IO_ERROR;
	}
	if(out_file && outfd>=0)
	{
		logp("do not specify both output file and output fd in do_rs_run()\n");
		return RS_IO_ERROR;
	}

	if((bfd || in_file || in_zfile || infd>=0)
	 && !(in_fb=rs_filebuf_new(bfd, in_file, in_zfile, infd, inz, rs_inbuflen)))
		return RS_MEM_ERROR;
	if((out_file || out_zfile || outfd>=0)
	 && !(out_fb=rs_filebuf_new(NULL, out_file, out_zfile, outfd, outz, rs_outbuflen)))
	{
		if(in_fb) rs_filebuf_free(in_fb);
		return RS_MEM_ERROR;
	}

//logp("before rs_job_drive\n");
	result = rs_job_drive(job, &buf,
		in_fb ? rs_infilebuf_fill : NULL, in_fb,
		out_fb ? rs_outfilebuf_drain : NULL, out_fb);
//logp("after rs_job_drive\n");

	// Do not worry about flushing zlib deflate - nobody uses that
	// with do_rs_run.

	if(in_fb) rs_filebuf_free(in_fb);
	if(out_fb) rs_filebuf_free(out_fb);

	return result;
}






static rs_result rs_async_drive(rs_job_t *job, rs_buffers_t *rsbuf,
             rs_driven_cb in_cb, void *in_opaque,
             rs_driven_cb out_cb, void *out_opaque)
{
	rs_result result;
	rs_result iores;

	if (!rsbuf->eof_in && in_cb)
	{
		iores = in_cb(job, rsbuf, in_opaque);
		if (iores != RS_DONE) return iores;
	}

	result = rs_job_iter(job, rsbuf);
	if (result != RS_DONE && result != RS_BLOCKED)
		return result;
	//printf("job iter got: %d\n", result);

	if (out_cb)
	{
    		rs_filebuf_t *fb = (rs_filebuf_t *) out_opaque;
		iores = (out_cb)(job, rsbuf, out_opaque);
		if (iores != RS_DONE) return iores;

		// May need to flush the zlib stuff.

		if(result==RS_DONE
		  && rsbuf->eof_in && fb && fb->zstrm && fb->z==ZTYPE_DEFLATE)
		{
			size_t wlen=0;
			iores=zlib_fd_deflate(fb, &wlen, Z_FINISH);
			//logp("last deflate: %d\n", iores);
			if(iores!=RS_DONE) return iores;
			//return iores;
		}
	}

	return result;
}

rs_result rs_async(rs_job_t *job, rs_buffers_t *rsbuf, rs_filebuf_t *infb, rs_filebuf_t *outfb)
{
	return rs_async_drive(job, rsbuf,
		infb ? rs_infilebuf_fill : NULL, infb,
		outfb ? rs_outfilebuf_drain : NULL, outfb);
}


static rs_result
rs_whole_gzrun(rs_job_t *job, gzFile in_zfile, FILE *out_file, gzFile out_zfile)
{
    rs_buffers_t    buf;
    rs_result       result;
    rs_filebuf_t    *in_fb = NULL, *out_fb = NULL;

    if (in_zfile)
        in_fb = rs_filebuf_new(NULL, NULL, in_zfile, -1, ZTYPE_NONE, rs_inbuflen);

    if (out_file || out_zfile)
        out_fb = rs_filebuf_new(NULL, out_file, out_zfile, -1, ZTYPE_NONE, rs_outbuflen);
//logp("before drive\n");
    result = rs_job_drive(job, &buf,
                          in_fb ? rs_infilebuf_fill : NULL, in_fb,
                          out_fb ? rs_outfilebuf_drain : NULL, out_fb);
//logp("after drive\n");

    if (in_fb)
        rs_filebuf_free(in_fb);

    if (out_fb)
        rs_filebuf_free(out_fb);

    return result;
}

rs_result rs_patch_gzfile(FILE *basis_zfile, gzFile delta_zfile, FILE *new_file, gzFile new_zfile, rs_stats_t *stats)
{
	rs_job_t            *job;
	rs_result           r;

	job = rs_patch_begin(rs_file_copy_cb, basis_zfile);

	r = rs_whole_gzrun(job, delta_zfile, new_file, new_zfile);
/*
	if (stats)
		memcpy(stats, &job->stats, sizeof *stats);
*/

	rs_job_free(job);

	return r;
}


rs_result rs_sig_gzfile(gzFile old_file, FILE *sig_file, size_t new_block_len, size_t strong_len, rs_stats_t *stats)
{
    rs_job_t        *job;
    rs_result       r;

    job = rs_sig_begin(new_block_len, strong_len);
    r = rs_whole_gzrun(job, old_file, sig_file, NULL);
/*
    if (stats)
        memcpy(stats, &job->stats, sizeof *stats);
*/
    rs_job_free(job);

    return r;
}

rs_result rs_loadsig_gzfile(gzFile sig_file, rs_signature_t **sumset, rs_stats_t *stats)
{
    rs_job_t            *job;
    rs_result           r;

    job = rs_loadsig_begin(sumset);
    r = rs_whole_gzrun(job, sig_file, NULL, NULL);
/*
    if (stats)
        memcpy(stats, &job->stats, sizeof *stats);
*/
    rs_job_free(job);

    return r;
}

rs_result rs_delta_gzfile(rs_signature_t *sig, gzFile new_file, gzFile delta_file, rs_stats_t *stats)
{
    rs_job_t            *job;
    rs_result           r;

    job = rs_delta_begin(sig);

    r = rs_whole_gzrun(job, new_file, NULL, delta_file);
/*
    if (stats)
        memcpy(stats, &job->stats, sizeof *stats);
*/

    rs_job_free(job);

    return r;
}












rs_result zlib_fd_deflate(rs_filebuf_t *fb, size_t *wlen, int flush)
{
	int zret=Z_OK;
	z_stream *zstrm=fb->zstrm;
//logp("in deflate: %d\n", *wlen);
/*
	static FILE *xp=NULL;
	static FILE *yp=NULL;
	char p[64]="";
	snprintf(p, sizeof(p), "/tmp/test.beforedef.%d", getpid());
	if(!xp) xp=fopen(p, "w");
	snprintf(p, sizeof(p), "/tmp/test.afterdef.%d", getpid());
	if(!yp) yp=fopen(p, "w");
*/

	if(fb->zhave)
	{
//logp("TRY TO WRITE A: %d\n", fb->zhave);
		// Still some left to write from a previous try. 
		//if(yp) fwrite(fb->zbuf, 1, fb->zhave, yp);
		if(async_append_all_to_write_buffer('a',
			(const char *)fb->zbuf, &(fb->zhave)))
				return RS_BLOCKED;
	}

	if(!zstrm->avail_out || flush==Z_FINISH)
	{
//logp("more to get out %d %d\n", zstrm->avail_out, flush);
		// Still have more to get out of deflate.
		zstrm->avail_out=ZCHUNK;
		zstrm->next_out=fb->zbuf;
		zret=deflate(zstrm, flush);
		if(zret==Z_STREAM_ERROR)
		{
			logp("error in zlib_fd_deflate\n");
			return RS_IO_ERROR;
		}
		if((fb->zhave=ZCHUNK-zstrm->avail_out))
		{
//logp("TRY TO WRITE B: %d\n", fb->zhave);
			//if(yp) fwrite(fb->zbuf, 1, fb->zhave, yp);
			if(async_append_all_to_write_buffer('a',
				(const char *)fb->zbuf, &(fb->zhave)))
					return RS_BLOCKED;
		}
		if(!zstrm->avail_out) return RS_BLOCKED;
	}

	if(*wlen)
	{
		static unsigned char *in=(unsigned char *)malloc(ZCHUNK);
		memcpy(in, fb->buf, *wlen);
		//if(xp) fwrite(fb->buf, 1, *wlen, xp);
		//zstrm->next_in=(unsigned char *)fb->buf;
		zstrm->next_in=in;
		zstrm->avail_in=*wlen;

		zstrm->avail_out=ZCHUNK;
		zstrm->next_out=fb->zbuf;
		zret=deflate(zstrm, flush);
		if(zret==Z_STREAM_ERROR)
		{
			logp("error in zlib_fd_deflate\n");
			return RS_IO_ERROR;
		}
                if((fb->zhave=ZCHUNK-zstrm->avail_out))
		{
			//logp("TRY TO WRITE C: %d\n", fb->zhave);
			//if(yp) fwrite(fb->zbuf, 1, fb->zhave, yp);
			if(async_append_all_to_write_buffer('a',
				(const char *)fb->zbuf, &(fb->zhave)))
					return RS_BLOCKED;
		}

		if(!zstrm->avail_out) return RS_BLOCKED;
		return RS_DONE;
	}

	if(flush==Z_FINISH && zret!=Z_STREAM_END) return RS_BLOCKED;

	return RS_DONE;
}



rs_result zlib_fd_inflate(int fd, rs_filebuf_t *fb, rs_buffers_t *buf)
{
	char rcmd;
	size_t rlen;
	char *rbuf=NULL;
	z_stream *zstrm=fb->zstrm;
	static int zret=Z_OK;
	static int goteof=0;

//logp("in inflate\n");

//	static FILE *xp=NULL;
//	char p[64]="";
//	snprintf(p, sizeof(p), "/tmp/test.del.%d", getpid());
//	if(!xp) xp=fopen(p, "w");

	if(goteof && zret==Z_STREAM_END)
	{
		buf->eof_in=1;
		zret=Z_OK;
		goteof=0;
		return RS_DONE;
	}

	if(!zstrm->avail_out)
	{
		unsigned char out[ZCHUNK];

		zstrm->avail_out=ZCHUNK;
		zstrm->next_out=out;
		zret=inflate(zstrm, Z_NO_FLUSH);
		switch(zret) {
			case Z_NEED_DICT:
			case Z_DATA_ERROR:
			case Z_MEM_ERROR:
			case Z_STREAM_ERROR:
				logp("error in zlib_fd_inflate: %d\n", zret);
				return RS_IO_ERROR;
		}
		buf->avail_in=ZCHUNK-zstrm->avail_out;
		memcpy(fb->buf, out, buf->avail_in);
		buf->next_in = fb->buf;

		//logp("inflate got a: %d (%d)\n", buf->avail_in, zret);
		//if(xp) fwrite(fb->buf, 1, buf->avail_in, xp);

		return RS_DONE;
	}

	if(async_read(&rcmd, &rbuf, &rlen)) return RS_IO_ERROR;
	if(rcmd=='a')
	{
		unsigned char out[ZCHUNK];
		//logp("got 'a' in zlib_fd_inflate: %d\n", rlen);

		if(!rlen)
		{
			logp("got 'a' with no data in zlib_fd_inflate\n");
			return RS_IO_ERROR;
		}
		memcpy(fb->zbuf, rbuf, rlen);
		free(rbuf);

		zstrm->avail_in=rlen;
		zstrm->next_in=fb->zbuf;


		zstrm->avail_out=ZCHUNK;
		zstrm->next_out=out;
		zret=inflate(zstrm, Z_NO_FLUSH);
		switch(zret) {
			case Z_NEED_DICT:
			case Z_DATA_ERROR:
			case Z_MEM_ERROR:
			case Z_STREAM_ERROR:
				logp("error in zlib_fd_inflate: %d\n", zret);
				return RS_IO_ERROR;
		}

		buf->avail_in = ZCHUNK-zstrm->avail_out;
		memcpy(fb->buf, out, buf->avail_in);
		buf->next_in = fb->buf;
	//	logp("inflate got b: %d (%d)\n", buf->avail_in, zret);
		//if(xp) fwrite(fb->buf, 1, buf->avail_in, xp);

		return RS_DONE;
	}
	else if(rcmd=='x')
	{
		free(rbuf);
		//logp("got 'x' in zlib_fd_inflate\n");
		goteof++;
		return RS_DONE;
	}
	else
	{
		logp("unexpected cmd in zlib_fd_inflate: %c:%s\n", rcmd, rbuf);
		free(rbuf);
		return RS_IO_ERROR;
	}

	return RS_DONE;
}

/*

rs_result zlib_fd_deflate(rs_filebuf_t *fb, size_t *wlen, int flush)
{
	int zret=Z_OK;
	z_stream *zstrm=fb->zstrm;

	static unsigned char *in=(unsigned char *)malloc(ZCHUNK);
	unsigned char out[ZCHUNK];

	if(*wlen)
	{
        	zstrm->avail_in = *wlen;
		memcpy(in, fb->buf, *wlen);
		zstrm->next_in = in;
	}
	do {
		zstrm->avail_out = ZCHUNK;
		zstrm->next_out = out;
		zret = deflate(zstrm, flush);
		if(zret==Z_STREAM_ERROR)
		{
			logp("error in zlib_fd_deflate\n");
			return RS_IO_ERROR;
		}
		if((fb->zhave = ZCHUNK - zstrm->avail_out))
			async_append_all_to_write_buffer('a', (const char *)out, &(fb->zhave));
        } while (!zstrm->avail_out);

	return RS_DONE;
}

rs_result zlib_fd_inflate(int fd, rs_filebuf_t *fb, rs_buffers_t *buf)
{
	static int zret=Z_OK;
	z_stream *zstrm=fb->zstrm;
	unsigned char in[ZCHUNK];
	unsigned char out[ZCHUNK];
	int goteof=0;
logp("enter inflate\n");

	do {
		size_t rlen;	
		char rcmd='\0';
		char *rbuf=NULL;
	logp("inflate loop a\n");
		if(!goteof && async_read(&rcmd, &rbuf, &rlen))
			return RS_IO_ERROR;
	logp("inflate loop x\n");
		if(rcmd=='x')
		{
			logp("got 'x'\n");
			free(rbuf);
			goteof=1;
		}
		else if(rcmd=='a' || goteof)
		{
			if(rlen && !goteof)
			{
				zstrm->avail_in = rlen;
				memcpy(in, rbuf, rlen);
				free(rbuf);
				zstrm->next_in = in;
			}
			do {
	logp("inflate loop b\n");
				zstrm->avail_out = ZCHUNK;
				zstrm->next_out = out;
				zret = inflate(zstrm, Z_NO_FLUSH);
				switch(zret) {
				case Z_NEED_DICT:
				case Z_DATA_ERROR:
				case Z_MEM_ERROR:
				case Z_STREAM_ERROR:
				  logp("error in zlib_fd_inflate: %d\n", zret);
				  return RS_IO_ERROR;
				}
				fb->zhave = ZCHUNK - zstrm->avail_out;
	logp("inflated: %d\n", fb->zhave);
				memcpy(fb->buf, out, fb->zhave);
				buf->next_in = fb->buf;
				buf->avail_in = fb->zhave;
			} while(!zstrm->avail_out);
		}
		else
		{
			logp("unexpected cmd in zlib_fd_inflate: %c:%s\n", rcmd, rbuf);
			free(rbuf);
			return RS_IO_ERROR;
		}
	} while(zret!=Z_STREAM_END);

	if(zret==Z_STREAM_END)
	{
		buf->eof_in=1;
		return RS_DONE;
	}
	return RS_IO_ERROR;
}
*/
